package com.topcoder.persistence.relation;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * This is a special concrete implementation of a relation. A relation consists of a tuple type which
 * indicates the attribute names and types a tuple belonging to the relation can have,
 * a set of tuple constraints indicating the conditions a tuple has to respect to be
 * a part of the tuple and a set of relation constraints indicating global conditions
 * that the entire set of tuples in the relation must respect.
 *
 * This class extends DefaultRelation. The only new functionality that it brings is lookup by key.
 * The class keeps an index, updated at each add, remove, clear operation, that allow fast O(1)
 * tuple finding by key. The findTuples method recognize if the find refers to the indexed
 * key attributes and uses the index to find the tuple. In contrast, the findTuples methods
 * from the super class have to iterate each tuple to find the result (O(n) complexity where n
 * is the number of tuples in the relation).
 * 
 * The user is supposed to give in the constructor the attributes the relation should index on.
 * 
 * Thread safety:
 * 
 * The class is not thread safe because it has state that can mutate.
 * There are lots of modifiers methods and many methods need access to several attributes.
 * That's why synchronizing all methods is the only reasonable solution if thread safety
 * is desired.
 */
public class IndexedRelation extends DefaultRelation {
	
	/**
	 * The key attributes to index by.
	 * Cannot be null, empty or contain nulls, empty or all spaces strings.
	 * Initialized in the constructor, used in all the methods.
	 */
	private String[] keyAttributes = null;
	
	/**
	 * The index on the key attributes.
	 * It maps Lists to Tuples or list of Tuples.
	 * Cannot be null, or have null keys or values.
	 * Used by all methods.
	 */
	private Map index = new HashMap();
	
	/**
	 * Constructor with no tuple type and one key attribute.
	 * 
	 * Implementation details:
	 * 
	 * - calls super(name)
	 * - copies keyAttribute to keyAttributes (creates an array)
	 * 
	 * @param name the name of the relation
	 * @param keyAttribute the key attribute to index on
	 * 
	 * @throws IllegalArgumentException if any string is null, empty or all spaces
	 */
	public IndexedRelation(String name, String keyAttribute) throws IllegalArgumentException {
		super(name);
	}

	/**
	 * Constructor with no tuple type and more key attributes.
	 * 
	 * Implementation details:
	 * 
	 * - calls super(name)
	 * - copies keyAttributes param to keyAttributes attributes (creates an array)
	 * 
	 * @param name the name of the relation
	 * @param keyAttributes the key attributes to index on
	 * 
	 * @throws IllegalArgumentException if any string or element is null, empty or all spaces
	 */
	public IndexedRelation(String name, String[] keyAttributes) throws IllegalArgumentException {
		super(name);
	}

	/**
	 * Constructor with tuple type and one key attribute.
	 * 
	 * Implementation details:
	 * 
	 * - calls super(name, tupleType)
	 * - copies keyAttribute to keyAttributes (creates an array)
	 * 
	 * @param name the name of the relation
	 * @param tupleType the type of the tuple
	 * @param keyAttribute the key attribute to index on
	 * 
	 * @throws IllegalArgumentException if any string is null, empty or all spaces
	 */
	public IndexedRelation(String name, TupleType tupleType, String keyAttribute) throws IllegalArgumentException {
		super(name, tupleType);
	}

	/**
	 * Constructor with tuple type and more key attributes.
	 * 
	 * Implementation details:
	 * 
	 * - calls super(name, tupleType)
	 * - copies keyAttributes param to keyAttributes attribute (creates an array)
	 * 
	 * @param name the name of the relation
	 * @param tupleType the type of the tuple
	 * @param keyAttributes the key attributes to index on
	 * 
	 * @throws IllegalArgumentException if any string or element is null, empty or all spaces
	 */
	public IndexedRelation(String name, TupleType tupleType, String[] keyAttributes) throws IllegalArgumentException {
		super(name, tupleType);
	}
	
	/**
	 * Adds a tuple to the relation.
	 * If the tuple already exists, nothing is done.
	 * 
	 * All tuple constraints are evaluated for the tuple to see if it meets them.
	 * If checkRelationConstraints == true, all relation constraints are checked too.
	 * If any constraint fails, an exception is thrown and the tuple is not added.
	 * 
	 * Implementation details:
	 * 
	 * - calls super.addTuple
	 * - builds a key ArrayList with the values of the key attributes
	 * - if the index map doesn't contain key list, put key list to tuple mapping
	 * - if the index map has a Tuple for the key list, replace it with an
	 *   ArrayList containing both tuples
	 * - if the index map has a Tuple list for the key list, add the tuple to the list
	 *  
	 * @param tuple the tuple to add
	 * 
	 * @return true if added, false if not (duplicate)
	 * 
	 * @throws IllegalArgumentException if the argument is null
	 * @throws EvaluationException in an error occurs while evaluating the constraint
	 * @throws ConstraintViolatedException if the relation violates a constraint
	 * @throws IllegalStateException if the tuple type has not been set yet
	 */
	public boolean addTuple(Tuple tuple) throws IllegalArgumentException, EvaluationException, ConstraintViolatedException, IllegalStateException {
		return false;
	}

	/**
	 * Removes a tuple from the relation.
	 * If the tuple doesn't exists, nothing is done.
	 * 
	 * If checkRelationConstraints == true, all relation constraints are checked.
	 * If any constraint fails, an exception is thrown and the tuple is not removed.
	 * 
	 * Implementation details:
	 * 
	 * - calls super.removeTuple
	 * - builds a key ArrayList with the values of the key attributes
	 * - if the index map has a Tuple for the key list, remove the entry
	 * - if the index map has a Tuple list for the key list, remove the tuple from the list
	 * 
	 * @param tuple the tuple to remove
	 * 
	 * @return true if removed, false if not (not found)
	 * 
	 * @throws IllegalArgumentException if the argument is null
	 * @throws EvaluationException in an error occurs while evaluating the constraint
	 * @throws ConstraintViolatedException if the relation violates a constraint
	 */
	public boolean removeTuple(Tuple tuple) throws IllegalArgumentException, EvaluationException, ConstraintViolatedException {
		return false;
	}
	
	/**
	 * Removes all tuples from the relation.
	 * 
	 * If checkRelationConstraints == true, all relation constraints are checked.
	 * If any constraint fails, an exception is thrown and the clear is not performed.
	 * 
	 * Implementation details:
	 * 
	 * - calls super.clearTuples
	 * - clears index map
	 * 
	 * @throws EvaluationException in an error occurs while evaluating the constraint
	 * @throws ConstraintViolatedException if the relation violates a constraint
	 */
	public void clearTuples() throws EvaluationException, ConstraintViolatedException {
	}	
	
    /**
     * Basic find method that finds tuples by an attribute value.
     * For anything that is more complex, probably the restrict operation is a better choice.
     * 
	 * Implementation details:
	 * 
	 * - checks if attribute matches key attributes array (should have one element,
	 *   equal with attribute)
	 * - if no, call super.findTuples
	 * - if yes, create key ArrayList with value and get the result from the index map by that key
     * 
     * @param attribute the attribute to find by
     * @param value the value of the attribute
     * 
     * @return the found tuples
     * 
     * @throws IllegalArgumentException if any argument is null or if the string is empty or all spaces
     */	
	public Tuple[] findTuples(String attribute, Object value) throws IllegalArgumentException {
		return null;
	}
	
    /**
     * Basic find method that finds tuples by some attribute values.
     * For anything that is more complex, probably the restrict operation is a better choice.
     * 
	 * Implementation details:
	 * 
	 * - checks if attributes matches key attributes array
	 * - if no, call super.findTuples
	 * - if yes, create key ArrayList with values and get the result from the index map by that key
     * 
     * @param attributes the attributes to find by
     * @param values the values of the attributes
     * 
     * @return the found tuples
     * 
     * @throws IllegalArgumentException if any argument is null, if the arrays have different size or
     * if they have null elements, empty or all spaces strings
     */	
	public Tuple[] findTuples(String attributes[], Object values[]) throws IllegalArgumentException {
		return null;
	}

	/**
	 * Notifier method called by Tuple when a value for a field changes.
	 * This gives the chance to enforce the tuple and relation constraints.
	 * 
	 * Implementation details:
	 * 
	 * - calls super.tupleChanged
	 * - create a key ArrayList with the key attribute values (using oldValue)
	 * - remove the old tuple under that key (see removeTuple)
	 * - create a key ArrayList with the key attribute values
	 * - add the new tuple under that key (see addTuple)
	 * 
	 * @param tuple the tuple (changed)
	 * @param attribute the attribute that has changed
	 * @param oldValue the old value of the attribute
	 * 
	 * @throws IllegalArgumentException if any argument is null or if the string is empty or all spaces
	 * @throws EvaluationException in an error occurs while evaluating the constraint
	 * @throws ConstraintViolatedException if the relation violates a constraint
	 */
	void tupleChanged(Tuple tuple, String attribute, Object oldValue) throws IllegalArgumentException, EvaluationException, ConstraintViolatedException {	
	}

}
