package com.topcoder.persistence.relation.constraints;

import java.util.HashSet;
import java.util.Set;

import com.topcoder.persistence.relation.EvaluationException;
import com.topcoder.persistence.relation.Relation;
import com.topcoder.persistence.relation.StatefulRelationConstraint;
import com.topcoder.persistence.relation.Tuple;

/**
 * This class implements a relation constraint that allows specifying
 * that an attribute or a combination of attributes must have unique
 * values.
 * 
 * The unique constraint is implemented as a stateful relation constraint.
 * Determining if the relation has unique values for each tuple addition,
 * removal or change is too time consuming (requires iterating all tuples).
 * A Set will all the currently used values is kept instead. This Set
 * is updated when tuples are added, removed or changed. 
 * 
 * Thread safety:
 * 
 * This class is not thread safe because it has state.
 * An instance of this class cannot be shared for more relations.
 * In a thread safe implementation the access to the usedValues set
 * would have to be synchronized using blocks. 
 * 
 * @author TCSDESIGNER
 */
public class UniqueConstraint implements StatefulRelationConstraint {
	
	/**
	 * The names of the attributes that form the key that must be unique.
	 * Cannot be null, have null elements, be empty.
	 */
	private String[] keyNames = null;
	
	/**
	 * A set with the currently used key values.
	 * It contains Lists.
	 * Each list contains the values for the key attributes
	 * Cannot be null or have null as elements.
	 * Modified by all methods. 
	 */
	private Set usedValues = new HashSet();
	
	/**
	 * Constructor.
	 * 
	 * keyNames = new String[] {keyName}
	 * 
	 * @param keyName the unique field the constraint checks
	 * 
	 * @throws IllegalArgumentException if the argument is null, empty or all spaces
	 */
	public UniqueConstraint(String keyName) throws IllegalArgumentException {
	}

	/**
	 * Constructor.
	 * 
	 * Copies the param to the attribute (defensive copy).
	 * 
	 * @param keyNames the unique field combination the constraint checks
	 * 
	 * @throws IllegalArgumentException if the argument is null, has null, empty or all spaces string elements
	 */
	public UniqueConstraint(String[] keyNames) throws IllegalArgumentException {
	}
	
	/**
	 * Getter for the key attribute names.
	 * The keyNames attribute is returned (defensive copy).
	 * 
	 * @return the key attribute names
	 */
	public String[] getKeyNames() {
		return null;
	}

	/**
	 * Checks whether the relation respects the constraint of having unique values
	 * in the key attributes.
	 * 
	 * Implementation details:
	 * 
	 * - clear usedValues
	 * - iterate tuples
	 * - for each tuple, put the values for the key attributes in an ArrayList
	 * - put the list in usedValues (if duplicate found then return false)
	 * - return true
	 * 
	 * @param relation the relation to check
	 * 
	 * @return true if constraint is met, false otherwise
	 * 
	 * @throws IllegalArgumentException if the argument is null
	 * @throws EvaluationException thrown if the evaluation cannot be done because of an
	 * implementation specific error 
	 */
	public boolean validate(Relation relation) throws IllegalArgumentException, EvaluationException {
		return false;
	}
	
	/**
	 * Update the usedValues set with a newly added tuple.
	 * 
	 * Implementation details:
	 * 
	 * - put the values for the key attributes in an ArrayList
	 * - put the list in usedValues (if duplicate found then return false)
	 * 
	 * @param tuple the added tuple
	 * 
	 * @return true if constraint is met, false otherwise
	 * 
	 * @throws IllegalArgumentException if the argument is null
	 * @throws EvaluationException thrown if the evaluation cannot be done because of an
	 * implementation specific error 
	 */
	public boolean tupleAdded(Tuple tuple) throws IllegalArgumentException, EvaluationException {
		return false;
	}

	/**
	 * Update the usedValues set when a tuple is removed.
	 * 
	 * Implementation details:
	 * 
	 * - put the values for the key attributes in an ArrayList
	 * - remove the list from usedValues
	 * - return true always
	 * 
	 * @param tuple the removed tuple
	 * 
	 * @return true always
	 * 
	 * @throws IllegalArgumentException if the argument is null
	 * @throws EvaluationException thrown if the evaluation cannot be done because of an
	 * implementation specific error 
	 */
	public boolean tupleRemoved(Tuple tuple) throws IllegalArgumentException, EvaluationException {
		return false;
	}

	/**
	 * Update the usedValues set when a tuple value is changed.
	 * 
	 * Implementation details:
	 * 
	 * - put the values for the key attributes in an ArrayList (use oldValue)
	 * - remove the list from usedValues
	 * - put the values for the key attributes in an ArrayList
	 * - add the list to usedValues (return false if duplicate exists)
	 * - return true always
	 * 
	 * @param tuple the changed tuple
	 * @param attribute the attribute that has changed
	 * @param oldValue the old value for the changed attribute
	 * 
	 * @return true if constraint is met, false otherwise
	 * 
	 * @throws IllegalArgumentException if the argument is null
	 * @throws EvaluationException thrown if the evaluation cannot be done because of an
	 * implementation specific error 
	 */
	public boolean tupleChanged(Tuple tuple, String attribute, Object oldValue) throws IllegalArgumentException, EvaluationException {
		return false;
	}

}
