package com.topcoder.persistence.relation.operations;

import com.topcoder.persistence.relation.OperationException;
import com.topcoder.persistence.relation.Relation;
import com.topcoder.persistence.relation.Tuple;
import com.topcoder.persistence.relation.TupleConstraint;

/**
 * Join implementation that uses as match criteria any tuple constraint.
 * When two tuples need to be tested for matching, they are combined into one
 * tuple (each name is qualified with "left." and "right." regardless of duplication).
 * The constraint is called for this combined tuple.
 * 
 * This type of join takes as input the tuple constraint to be used for matching.
 * The result will be obtained from all the tuple pairs from the two relations that
 * respect the constraint.
 * 
 * In addition, aliases are also suported.
 * Aliases are used when two relations have common attribute names.
 * Joining them would result in name conflict so aliasing allows
 * renaming fields in order to avoid that.
 * 
 * Outer joins are supported by keeping a special tuple, the missingValues
 * tuple. When a tuple from the left relation doesn't have a match in the
 * right relation, the left tuple is combined with the special missingValues
 * tuple. This approach is chosen to respect the requirement saying nulls
 * should not be used (the most natural value to be used for missing values).
 * 
 * The result relation is given by the user (must be empty and typeless).
 * This allows the user to select the desired relation implementation
 * (optimized for speed, for memory, etc).
 * 
 * Thread safety:
 * 
 * This class is immutable so it is thread safe.
 * 
 * @author TCSDESIGNER
 */
public class ConditionalJoinOperation extends JoinOperation {
	
	/**
	 * The tuple constraint that is used to determine whether two tuples match.
	 * When two tuples need to be tested for matching, they are combined into one
	 * tuple (each name is qualified with "left." and "right." regardless of duplication).
	 * The constraint is called for this combined tuple.
	 *    
	 * Can be null (means everything matches everything and it is used only to support the product operation).
	 * Set in the constructor used in compute.
	 */
	private TupleConstraint constraint = null;
		
	/**
	 * Protected constructor to be used by product operation (subclass).
	 * They are conditional joins with no match constraint.
	 */
	protected ConditionalJoinOperation() {
	}
	
	/**
	 * Protected constructor to be used by product operation (subclass).
	 * They are conditional joins with no match constraint.
	 * This overload allows specifying aliases.
	 * 
	 * Implementation details:
	 * 
	 * Calls super to handle aliases.
	 * 
     * @param leftAliased the attributes from the left relation that are aliased 
     * @param leftAliases the aliases for the aliased attributes from the left table
     * @param rightAliased the attributes from the right relation that are aliased
     * @param rightAliases the aliases for the aliased attributes from the right table
     * 
     * @throws IllegalArgumentException if any argument is null, has null elements or 
	 * empty or all spaces strings or if the sizes of the pair arrays are different
	 */
	protected ConditionalJoinOperation(String[] leftAliased, String[] leftAliases, String[] rightAliased, String[] rightAliases) throws IllegalArgumentException {
	}

	/**
	 * Constructor specifying the join constraint.
	 * This will construct a normal (inner) join.
	 * 
	 * Implementation details:
	 * 
	 * Copies the parameter to the corresponding attribute.
	 * 
	 * @param constraint the tuple constraint that is used to determine whether two tuples match
	 * 
	 * @throws IllegalArgumentException if the argument is null
	 */
	public ConditionalJoinOperation(TupleConstraint constraint) throws IllegalArgumentException {
	}

	/**
	 * Constructor specifying the join constraint and the aliases.
	 * This will construct a normal (inner) join.
	 * 
	 * Implementation details:
	 * 
	 * Copies the constraint parameter to the corresponding attribute.
	 * Calls super to handle aliases.
	 * 
	 * @param constraint the tuple constraint that is used to determine whether two tuples match
     * @param leftAliased the attributes from the left relation that are aliased 
     * @param leftAliases the aliases for the aliased attributes from the left table
     * @param rightAliased the attributes from the right relation that are aliased
     * @param rightAliases the aliases for the aliased attributes from the right table
     * 
     * @throws IllegalArgumentException if any argument is null, has null elements or 
	 * empty or all spaces strings or if the sizes of the pair arrays are different
     */
	public ConditionalJoinOperation(TupleConstraint constraint, String[] leftAliased, String[] leftAliases, String[] rightAliased, String[] rightAliases) throws IllegalArgumentException {
		super(leftAliased, leftAliases, rightAliased, rightAliases);
	}

	/**
	 * Constructor specifying the join constraint.
	 * This will construct an outer join (a tuple with the missing values is given too).
	 * 
	 * Implementation details:
	 * 
	 * Copies the constraint parameter to the corresponding attribute.
	 * Calls super to handle missingValues.
	 * 
	 * @param constraint the tuple constraint that is used to determine whether two tuples match
     * @param missingValues the tuple specifying the missing values to be used for outer joins
     * 
     * @throws IllegalArgumentException if any argument is null
     */
	public ConditionalJoinOperation(TupleConstraint constraint, Tuple missingValues) throws IllegalArgumentException {
		super(missingValues);
	}

	/**
	 * Constructor specifying the join constraint and the aliases.
	 * This will construct an outer join (a tuple with the missing values is given too).
	 * 
	 * Implementation details:
	 * 
	 * Copies the constraint parameter to the corresponding attribute.
	 * Calls super to handle aliases and missing values.
	 * 
	 * @param constraint the tuple constraint that is used to determine whether two tuples match
     * @param leftAliased the attributes from the left relation that are aliased 
     * @param leftAliases the aliases for the aliased attributes from the left table
     * @param rightAliased the attributes from the right relation that are aliased
     * @param rightAliases the aliases for the aliased attributes from the right table
     * @param missingValues the tuple specifying the missing values to be used for outer joins
     * 
     * @throws IllegalArgumentException if any argument is null, has null elements or 
	 * empty or all spaces strings or if the sizes of the pair arrays are different
     */
	public ConditionalJoinOperation(TupleConstraint constraint, String[] leftAliased, String[] leftAliases, String[] rightAliased, String[] rightAliases, Tuple missingValues) throws IllegalArgumentException {
		super(leftAliased, leftAliases, rightAliased, rightAliases, missingValues);
	}

	/**
	 * Returns the tuple constraint that is used to determine whether two tuples match.
	 * 
	 * Implementation details:
	 * 
	 * Returns the constraint attribute.
	 * 
	 * @return the tuple constraint that is used to determine whether two tuples match
	 */
	public TupleConstraint getConstraint() {
		return constraint;
	}

	/**
	 * Computes the join of two relations.
	 * The join criteria consists of a any tuple constraint.
	 * The join operation looks for pairs of tuples from the two relations for which the constraint
	 * evaluates to true and puts them in the result relation.
	 * 
	 * When two tuples need to be tested for matching, they are combined into one
	 * tuple (each name is qualified with "left." and "right." regardless of duplication).
	 * The constraint is evaluated on this combined tuple.
	 * 
	 * Because the relations may have common attribute names, the
	 * user can use aliases. If conflicts still remain, the names are qualified with "left." and "right."
	 * to avoid any confusion.
	 * 
	 * Note that if the missingValues attribute is not null, an outer join is performed.
	 * If no match if found for a tuple from the left relation, it is still put in the result
	 * relation but combined with the missingValues tuple.
	 * 
	 * Implementation details:
	 * 
	 * The join algorithm consists of two parts. The first part is solving attribute name
	 * conflicts in the output relation. The mechanism is simple: if the user specifies
	 * aliases, they attributes are renamed accordingly; if conflicts still remain they are 
	 * qualified with "left." and "right.".
	 * 
	 * The second step is iterating all pairs of tuples from the two relations (nested loops).
	 * If the constraint applied on a combined tuple (from the pair) is true, then we have a match
	 * so we add it to the result. When no match is found for a left tuple and missingValues is
	 * not null then it means we have an outer join so we combine the tuple with missingValues
	 * and add it to the result relation.
	 * 
	 * Part 1 processes the attribute names only so it is not affected by the number of tuples
	 * in the relation (in terms of algorithmic complexity). Part 2 has complexity O(m*n) because
	 * each pair of tuples is treated separately (m and n being the sizes of the relations).
	 * Therefore the global complexity is O(m*n). 
	 * 
	 * 1.1 - check if leftAttributes appear in leftRelation  
	 * 1.2 - check if rightAttributes appear in rightRelation
	 * 1.3 - create two attribute alias HashMaps for the two relationships
	 * 1.4 - if alias arrays not null, use them to populate the alias HashMaps 
	 * 1.5 - find all common attributes for the relations, even with aliases
	 * 1.6 - prefix the common attributes with "left." and "right." and put them
	 *       in the alias maps (for easier usage later)
	 * 1.7 - create a tuple type with all attributes (using the alias maps to obtain
	 *       the correct names)
	 * 1.8 - set the tuple type to the result relation
	 *  
	 * 2.1 - iterate the tuples from the left relation
	 * 2.2 - iterate the tuples from the right relation (nested loop)
	 * 2.2 - combine the two tuples into one tuple (all attributes are qualified with "left." and "right."
	 * 2.3 - evaluate the constraint and if true output tuple using the two matched tuples
	 * 2.4 - if no match for a left tuple and missingValues is not null, output left tuple combined with missingValues
	 *   
	 * 3.1 - output of a tuple means creating a tuple as shown below and adding it to the result relation
	 * 3.2 - create a new tuple
	 * 3.2 - iterate each attribute from the two tuples
	 * 3.3 - get the aliased name from the alias map if it exists, otherwise use the original name
	 * 3.4 - put the tuple value in the new tuple 
	 *  
	 * @param leftRelation the left relation
	 * @param rightRelation the right relation
	 * @param resultRelation the result relation (must be empty and typeless)
	 * 
	 * @throws IllegalArgumentException if any argument is null or the result relation is not empty and typeless
	 * @throws OperationException if any of the checks in the join algorithm fails
	 */
	public void compute(Relation leftRelation, Relation rightRelation, Relation resultRelation) throws IllegalArgumentException, OperationException {
	}
	
}
