package com.topcoder.persistence.relation.operations;

import com.topcoder.persistence.relation.OperationException;
import com.topcoder.persistence.relation.Relation;
import com.topcoder.persistence.relation.Tuple;

/**
 * Join implementation that uses as match criteria attribute value equality
 * for a set of attributes (equi-join).
 * 
 * This type of join takes as input the attribute names from the left and 
 * right relation that must match. The result will be obtained from all the
 * tuple pairs from the two relations that have equal attribute value for
 * the given join attributes.
 * 
 * In addition, aliases are also suported.
 * Aliases are used when two relations have common attribute names.
 * Joining them would result in name conflict so aliasing allows
 * renaming fields in order to avoid that.
 * 
 * Outer joins are supported by keeping a special tuple, the missingValues
 * tuple. When a tuple from the left relation doesn't have a match in the
 * right relation, the left tuple is combined with the special missingValues
 * tuple. This approach is chosen to respect the requirement saying nulls
 * should not be used (the most natural value to be used for missing values).
 *
 * The result relation is given by the user (must be empty and typeless).
 * This allows the user to select the desired relation implementation
 * (optimized for speed, for memory, etc).
 * 
 * Thread safety:
 * 
 * This class is immutable so it is thread safe.
 * 
 * @author TCSDESIGNER
 */
public class EqualsJoinOperation extends JoinOperation {
	
	/**
	 * The join attributes for the left relation.  
	 * Can be null but cannot contain nulls, empty or all spaces strings.
	 * Set in the constructor and with the setter and used in compute.
	 * leftAttributes and rightAttributes must both be null or both have
	 * the same size.
	 */
	private String[] leftAttributes = null;
	
	/**
	 * The join attributes for the right relation.  
	 * Can be null but cannot contain nulls, empty or all spaces strings.
	 * Set in the constructor and with the setter and used in compute.
	 * leftAttributes and rightAttributes must both be null or both have
	 * the same size.
	 */
	private String[] rightAttributes = null;
	
	/**
	 * Protected constructor to be used by natural joins (subclass).
	 * They don't know the join columns before compute is called.
	 * They will use setJoinAttributes to set them when they are known.
	 */
	protected EqualsJoinOperation() {
	}

	/**
	 * Constructor specifying the join attributes in the left and right relations.
	 * This will construct a normal (inner) join.
	 * 
	 * Implementation details:
	 * 
	 * Copies the parameters to the corresponding attributes (defensive array copies are made).
	 * 
	 * @param leftAttributes the left relation attributes for the join
	 * @param rightAttributes the right relation attributes for the join
	 * 
	 * @throws IllegalArgumentException if any argument is null, has null elements or 
	 * empty or all spaces strings or if the sizes of the arrays are different
	 */
	public EqualsJoinOperation(String[] leftAttributes, String[] rightAttributes) throws IllegalArgumentException {
	}

	/**
	 * Constructor specifying the join attributes in the left and right relations and the aliases.
	 * This will construct a normal (inner) join.
	 * 
	 * Implementation details:
	 * 
	 * Copies the parameters to the corresponding attributes (defensive array copies are made).
	 * Calls super to handle aliases.
	 * 
	 * @param leftAttributes the left relation attributes for the join
	 * @param rightAttributes the right relation attributes for the join
     * @param leftAliased the attributes from the left relation that are aliased 
     * @param leftAliases the aliases for the aliased attributes from the left table
     * @param rightAliased the attributes from the right relation that are aliased
     * @param rightAliases the aliases for the aliased attributes from the right table
     * 
     * @throws IllegalArgumentException if any argument is null, has null elements or 
	 * empty or all spaces strings or if the sizes of the pair arrays are different
     */
	public EqualsJoinOperation(String[] leftAttributes, String[] rightAttributes, String[] leftAliased, String[] leftAliases, String[] rightAliased, String[] rightAliases) throws IllegalArgumentException {
		super(leftAliased, leftAliases, rightAliased, rightAliases);
	}

	/**
	 * Constructor specifying the join attributes in the left and right relations.
	 * This will construct an outer join (a tuple with the missing values is given too).
	 * 
	 * Implementation details:
	 * 
	 * Copies the parameters to the corresponding attributes (defensive array copies are made).
	 * Calls super to handle missingValues.
	 * 
	 * @param leftAttributes the left relation attributes for the join
	 * @param rightAttributes the right relation attributes for the join
     * @param missingValues the tuple specifying the missing values to be used for outer joins
     * 
     * @throws IllegalArgumentException if any argument is null, has null elements or 
	 * empty or all spaces strings or if the sizes of the pair arrays are different
     */
	public EqualsJoinOperation(String[] leftAttributes, String[] rightAttributes, Tuple missingValues) throws IllegalArgumentException {
		super(missingValues);
	}

	/**
	 * Constructor specifying the join attributes in the left and right relations and the aliases.
	 * This will construct an outer join (a tuple with the missing values is given too).
	 * 
	 * Implementation details:
	 * 
	 * Copies the parameters to the corresponding attributes (defensive array copies are made).
	 * Calls super to handle aliases and missing values.
	 * 
	 * @param leftAttributes the left relation attributes for the join
	 * @param rightAttributes the right relation attributes for the join
     * @param leftAliased the attributes from the left relation that are aliased 
     * @param leftAliases the aliases for the aliased attributes from the left table
     * @param rightAliased the attributes from the right relation that are aliased
     * @param rightAliases the aliases for the aliased attributes from the right table
     * @param missingValues the tuple specifying the missing values to be used for outer joins
     * 
     * @throws IllegalArgumentException if any argument is null, has null elements or 
	 * empty or all spaces strings or if the sizes of the pair arrays are different
     */
	public EqualsJoinOperation(String[] leftAttributes, String[] rightAttributes, String[] leftAliased, String[] leftAliases, String[] rightAliased, String[] rightAliases, Tuple missingValues) throws IllegalArgumentException {
		super(leftAliased, leftAliases, rightAliased, rightAliases, missingValues);
	}

	/**
	 * Sets the join attributes.
	 * Useful for implementations that know the join attributes only when they see the relations 
	 * (natural join is an example of such join).
	 * 
	 * Implementation details:
	 * 
	 * Copies the parameters to the corresponding attributes (defensive array copies are made).
	 * 
	 * @param leftAttributes the left relation attributes for the join
	 * @param rightAttributes the right relation attributes for the join
     * 
     * @throws IllegalArgumentException if any argument is null, has null elements or 
	 * empty or all spaces strings or if the sizes of the pair arrays are different
	 */
	protected void setJoinAttributes(String[] leftAttributes, String[] rightAttributes) throws IllegalArgumentException {
	}

	/**
	 * Returns the join attributes for the left relation.
	 * 
	 * Implementation details:
	 * 
	 * Returns the leftAttributes attribute (defensive copy).
	 * 
	 * @return the join attributes for the left relation
	 */
	public String[] getLeftAttributes() {
		return leftAttributes;
	}

	/**
	 * Returns the join attributes for the right relation.
	 * 
	 * Implementation details:
	 * 
	 * Returns the rightAttributes attribute (defensive copy).
	 * 
	 * @return the join attributes for the right relation
	 */
	public String[] getRightAttributes() {
		return rightAttributes;
	}

	/**
	 * Computes the join of two relations.
	 * The join criteria consists of pairs of attributes from the two relations that must be equal.
	 * The join operation looks for pairs of tuples from the two relations for which all the values
	 * of the join attributes are equal and puts them in the result relation.
	 * 
	 * Because the relations may have common attribute names (that aren't involved in the join), the
	 * user can use aliases. If conflicts still remain, the names are qualified with "left." and "right."
	 * to avoid any confusion.
	 * 
	 * Note that if the missingValues attribute is not null, an outer join is performed.
	 * If no match if found for a tuple from the left relation, it is still put in the result
	 * relation but combined with the missingValues tuple.
	 * 
	 * Implementation details:
	 * 
	 * The join algorithm consists of three parts. The first part is solving attribute name
	 * conflicts in the output relation. The mechanism is simple: if the user specifies
	 * aliases, they attributes are renamed accordingly; if conflicts still remain in
	 * the non-joined attributes, they are qualified with "left." and "right.".
	 * 
	 * The second step is iterating of the right relation tuples and creating join keys
	 * from the join attributes (the key is an ArrayList). These keys are put in
	 * a HashMap, linked to the tuple it was obtained from. Since multiple tuples
	 * might have the same key, a key might be linked to several tuples.
	 * 
	 * The next step is iterating the left relation tuples. Join keys are created from the
	 * join attributes as in second step. By looking in the map obtained at step two under
	 * the join key, join tuple matches are found immediately (when key not found in the map,
	 * the tuple is combined with missingValues if not null - outer joins).
	 * 
	 * The matched tuple pairs are combined and a new tuple is created (using the aliases
	 * and qualification from step 1).
	 * 
	 * Part 1 processes the attribute names only so it is not affected by the number of tuples
	 * in the relation (in terms of algorithmic complexity). Part 2 has complexity O(n) because
	 * each tuple is added to the map in O(1); n is the size of the right relation.
	 * Part 3 is complicated. It is normally O(m), m being the size of the left relation,
	 * if the join attributes form a unique key in each relation. But if unique keys aren't
	 * formed, then it means one left tuple might match more right tuples. In the most extreme
	 * case, where all tuples have the same join key, it means everything matches everything,
	 * so we have m*n results (product). Obviously, the complexity cannot be smaller than the size of the
	 * output, therefore the complexity for part 3 is O(max(m,result_size)). That means
	 * globally the complexity is O(n + max(m, result_size)).
	 * 
	 * 1.1 - check if leftAttributes appear in leftRelation  
	 * 1.2 - check if rightAttributes appear in rightRelation
	 * 1.3 - create two attribute alias HashMaps for the two relationships
	 * 1.4 - if alias arrays not null, use them to populate the alias HashMaps 
	 * 1.5 - find all common attributes for the relations, even with aliases
	 * 1.6 - ignore those that are part of the join criteria (there is a position k
	 *       for which leftAttributes[k] == rightAttributes[k] == commonAttribute)
	 * 1.7 - prefix the common attributes with "left." and "right." and put them
	 *       in the alias maps (for easier usage later)
	 * 1.8 - create a tuple type with all attributes (using the alias maps to obtain
	 *       the correct names)
	 * 1.9 - set the tuple type to the result relation
	 *  
	 * 2.1 - iterate the tuples from the right relation
	 * 2.2 - for each tuple, create a key ArrayList with the values of the rightAttributes
	 * 2.3 - add to the map the (list, tuple) pair, if there is no entry for key list
	 * 2.4 - if key list is already mapped to another tuple, map it to an ArrayList containing both
	 * 2.5 - if key list is already mapped to a list of tuples, add the new tuple to the list
	 *   
	 * 3.1 - iterate the tuples from the left relation
	 * 3.2 - for each tuple, create a key ArrayList with the values of the leftAttributes
	 * 3.3 - get the value for the map entry for key list
	 * 3.4 - if the value is null and missingValues is not null (outer join), output tuple combined with missingValues
	 * 3.5 - if the value is a tuple, output iterated tuple combined with value tuple
	 * 3.6 - if the value is a list of tuples, output iterated tuple combined with each tuple in the list
	 * 
	 * 4.1 - output of a tuple means creating a tuple as shown below and adding it to the result relation
	 * 4.2 - create a new tuple
	 * 4.2 - iterate each attribute from the two tuples
	 * 4.3 - get the aliased name from the alias map if it exists, otherwise use the original name
	 * 4.4 - put the tuple value in the new tuple 
	 *  
	 * @param leftRelation the left relation
	 * @param rightRelation the right relation
	 * @param resultRelation the result relation (must be empty and typeless)
	 * 
	 * @throws IllegalArgumentException if any argument is null or the result relation is not empty and typeless
	 * @throws OperationException if any of the checks in the join algorithm fails
	 */
	public void compute(Relation leftRelation, Relation rightRelation, Relation resultRelation) throws IllegalArgumentException, OperationException {
	}
	
}
