package com.topcoder.persistence.relation.operations;

import com.topcoder.persistence.relation.BinaryOperation;
import com.topcoder.persistence.relation.IncompatibleRelationsException;
import com.topcoder.persistence.relation.Relation;

/**
 * Implementation of a binary operation for the intersection of two relations.
 * The result of this operation is a new relation that contains all the tuples
 * that appear in both relations.
 *
 * This operation applies only to relations with the same tuple type
 * (same attribute names, same attribute types). See TupleType.equals.
 * 
 * The result relation is given by the user (must be empty and typeless).
 * This allows the user to select the desired relation implementation
 * (optimized for speed, for memory, etc).
 * 
 * Note:
 * 
 * Theoretically an intersection is the same thing as a natural join
 * if the tuple types are equal (in this case the natural join will
 * try to match all attributes thus selecting the tuples that are present
 * in both sets). However, using the generic natural join algorithm
 * is too complicated for this simple intersection case.
 * 
 * Thread safety:
 * 
 * This class has no state so it is thread safe.
 * 
 * @author TCSDESIGNER
 */
public class IntersectionOperation implements BinaryOperation {

	/**
	 * Default constructor that doesn't do anything.
	 */
	public IntersectionOperation() {
	}

	/**
	 * Computes the intersection of the two relations.
	 * The result relation will contain all tuples that appear in both relations.
	 * 
	 * Implementation details:
	 * 
	 * The intersection algorithm is simply putting the tuples from the left
	 * relation in a HashSet, iterating the right relation tuples and adding
	 * to the result relation those tuples that are present in the set.
     * Each HashSet insertion and contains test is O(1) so the complexity is O(m+n)
	 * where m and n are the sizes of the two relations.
	 * 
	 * - check whether leftRelation.getTupleType().equals(rightRelation.getTupleType())
	 * - the tuple type of resultRelation is copied from any of the two relations
	 * - creates a new HashSet
	 * - calls leftRelation.getTuplesIterator()
	 * - iterates all tuples using the iterator and adds them to the set
	 * - calls rightRelation.getTuplesIterator()
	 * - iterates all tuples using the iterator and if present in the set, adds them to the result relation
	 * 
	 * @param leftRelation the left relation
	 * @param rightRelation the right relation
	 * @param resultRelation the result relation (must be empty and typeless)
	 * 
	 * @throws IllegalArgumentException if any argument is null or the result relation is not empty and typeless
	 * @throws IncompatibleRelationsException if the tuple types of the two relations are not equal
	 */
	public void compute(Relation leftRelation, Relation rightRelation, Relation resultRelation) throws IllegalArgumentException, IncompatibleRelationsException {
	}

}
