package com.topcoder.persistence.relation.operations;

import com.topcoder.persistence.relation.OperationException;
import com.topcoder.persistence.relation.Relation;
import com.topcoder.persistence.relation.UnaryOperation;

/**
 * Implementation of a unary operation for the projection of a relation
 * for a given list of attributes. The result of this operation is a new 
 * relation that keeps only the given projection attributes. Since by
 * removing some attributes two tuples might become equal, the projection
 * operation also removes duplicates in the result. 
 * 
 * The result relation is given by the user (must be empty and typeless).
 * This allows the user to select the desired relation implementation
 * (optimized for speed, for memory, etc).
 * 
 * Thread safety:
 * 
 * This class is immutable so it is thread safe.
 * 
 * @author TCSDESIGNER
 */
public class ProjectOperation implements UnaryOperation {
	
	/**
	 * The attributes the projection selects.
	 * Cannot be null, contain nulls, empty or all spaces strings.
	 * Set in the constructor and used in compute.
	 */
	private String[] attributes = null;
	
	/**
	 * Constructor.
	 * 
	 * Implementation details:
	 * 
	 * Sets the attributes attribute (makes a defensive copy).
	 * 
	 * @param attributes the attributes the projection selects
	 * 
	 * @throws IllegalArgumentException if the argument is null, contains nulls, 
	 * empty or all space strings or if it has duplicates
	 */
	public ProjectOperation(String[] attributes) throws IllegalArgumentException {
	}
	
	/**
	 * Returns the projection attributes.
	 * 
	 * Implementation details:
	 * 
	 * Returns the attributes attribute.
	 * 
	 * @return the projection attributes
	 */
	public String[] getAttributes() {
		return null;
	}

	/**
	 * Computes the projection of a source relation.
	 * The attribute list for the projection is given in the constructor. 
	 * The result relation will contain the subtuples of all tuples from 
	 * the source relation (subtuple meaning a tuple containing only the
	 * attributes given in the list).
	 * 
	 * Stripping some attributes might leave duplicate tuples so they are
	 * removed. 
	 * 
	 * Implementation details:
	 * 
	 * The projection algorithm is simply iterating all tuples from the source
	 * relation, creating subtuples from them and adding them to the result
	 * relation. A HashSet is used for duplicate detection. Each tuple is
	 * processed in O(1) and the HashSet insertions are O(1) too. Therefore
	 * the complexity is O(n) where n is the size of the source relation.
	 * 
	 * Note that adding the tuples directly to the result relation (no HashSet) may work
	 * for DefaultRelation (because that implementation is able to detect duplicates
	 * in O(1)) but might not work for other implementations (for which duplicate
	 * detection might not even be available because it would take too much time).
	 * The point is that the HashSet is needed. However, the developer is free
	 * to test with instanceof if the resultRelation is a DefaultRelation and
	 * not use the HashSet in that case. 
	 * 
	 * - check whether sourceRelation has all projection attributes 
	 * - creates a tuple type with the attributes and types of the projection
	 *   (the attributes are given in the constructor, the types are taken from the source relation)
	 * - sets the tuple type to the resultRelation
	 * - create a new HashSet (for duplicate detection)
	 * - calls sourceRelation.getTuplesIterator()
	 * - iterates all tuples using the iterator
	 * - creates a new tuple containing only the projection attributes
	 * - adds the tuple to the set 
	 * - adds all tuples from the set to the result relation
	 * 
	 * @param sourceRelation the source relation
	 * @param resultRelation the result relation (must be empty and typeless)
	 * 
	 * @throws IllegalArgumentException if any argument is null or the result relation is not empty and typeless
	 * @throws OperationException if the projection attributes cannot be found in the source relation  
	 */
	public void compute(Relation sourceRelation, Relation resultRelation) throws IllegalArgumentException, OperationException {
	}

}
